/* jshint undef: true, unused: true, esnext: true */
/* global plugin, require, Packages, com, java, sync, importPackage, Packages */
/* global Flag */

var facade = require("facade");
var Objects = require("convenience").Objects;
var _ = require("lodash");
var on = require("on")(plugin);
var FileUtil = Packages.jodd.io.FileUtil;
var StringUtilities = Packages.cattura.utils.string.StringUtilities;
var Executors = java.util.concurrent.Executors;
var MoreExecutors = com.google.common.util.concurrent.MoreExecutors;
var mediaProxy = facade.retrieveProxy("MediaProcessingProxy");
var environment = facade.retrieveProxy("EnvironmentProxy");
var devices = facade.retrieveProxy("DeviceProxy");
var Element = com.cattura.share.processor.data_provider.Element;
var Processor = com.cattura.share.processor.Processor;
var System = java.lang.System;
var File = java.io.File;
var LOGGER = plugin.info().logger;
var PPM_DIR = new File(environment.getStorageDirPath(), "ppm");
var TemplateHelper = Packages.cattura.agents.configuration.model.vos.capture.helpers.TemplateHelper;
var MediaTypes = Packages.cattura.utils.media.properties.MediaTypes;
var Runner = Packages.cattura.utils.concurrent.vos.Runner;
var IncidentReport = Packages.cattura.agents.configuration.model.vos.IncidentReport;
var FutureStatusWatcher = Packages.cattura.agents.configuration.model.vos.watchers.FutureStatusWatcher;
var Duration = java.time.Duration;
var ChronoUnit = java.time.temporal.ChronoUnit;
var Futures = com.google.common.util.concurrent.Futures;
var POLL_TIMEOUT_MS = 500; // Amount of time to await until dispatch new PPM data over event source [Milliseconds]
var WATCHER_TIMEOUT_SECONDS = 30; // Amount of time to wait for start/stop PPM session to complete, before sending an incident report [Seconds]
var CatturaEvent = Packages.cattura.agents.configuration.model.vos.events.Event;

var Maps = com.google.common.collect.Maps;
var Wildcard = Packages.jodd.util.Wildcard;

/*
 * Whether to kill PPM sessions when a recording starts up.
 * In v4 this was always true, in v5 it is false.
 * (16 October 2019: changed back to true to attempt to resolve sync problems)
 */
var STOP_PPM_SESSIONS_FOR_CAPTURE_SESSIONS = true;

importPackage(com.cattura.share.processor.data_provider.elements);

/*
 * Clear out PPM graphs directory
 */
if (PPM_DIR.exists())
{
    FileUtil.cleanDir(PPM_DIR);
}

function getConfiguredDeviceFilter ()
{
    var filters = Objects.toJS(plugin.systemSettings().get("filters").orElse([]), true);
    return device => !!_(filters).find(filter => Wildcard.equalsOrMatch(device.name, filter));
}

var PPM = (function() {
    var AUDIO_LEVEL_PLUGIN = "com.cattura.plugins.audio.analysis.level.AudioLevelPlugin";
    var ON_AUDIO_LEVEL_UPDATE_PLUGIN_SERVICE = "onAudioLevelUpdate";
    var EXECUTOR = MoreExecutors.listeningDecorator(Executors.newCachedThreadPool());
    var PPM_ONLY_CONFIG = {
        calculatePPM: new Flag(true),
        calculateMaxPeak: new Flag(false),
        calculateMinPeak: new Flag(false),
        calculateNoiseFloor: new Flag(false),
        calculateRMS: new Flag(false)
    };

    var _devices = []; // Detected devices.
    var _futureDevicePPMSessions = {}; // Device ID is used for key
    var _activeCaptureGraphs = {}; // MediaPackage ID is used for key
    var _ppmLevelsForGraph = {}; // Graph Root (as String) is used for key

    // Filters to exclude devices from PPM
    var deviceFilters = _(() => [
        device => !device.isUsable(),             // Filter unusable devices            
        device => device.isUSBDevice(),           // Filter USB devices
        device => device.name.startsWith("bmd:"), // Filter BMD devices (for now)
        getConfiguredDeviceFilter()               // Filters configurable in plugin system settings.
    ]).memoize().value();

    /*
     * Utility function to get PPM levels, for all devices, from a given graph
     *
     * @param {type} graph
     * @returns {main_L40.ppmLevelsForGraph|main_L40.getPPMLevelsFromGraph.mainAnonym$0}
     */
    function getPPMLevelsFromGraph ($graph)
    {
        return $graph.root.toString() in _ppmLevelsForGraph ? Objects.toJS(_ppmLevelsForGraph[$graph.root.toString()])
                                                            : {};
    }

    function makeAudioLevelsObject (key, audioMetrics)
    {
        return _({
            "@class": "com.cattura.share.processor.data_provider.elements.AudioLevel",
            id: null,
            name: null,
            tags: [],
            sourceID: key
        }).extend(audioMetrics).extend({
            timestamp: {
                "@class": "com.cattura.share.processor.data_provider.elements.Timestamp",
                id: null,
                name: null,
                tags: [],
                timestamp: audioMetrics.timestamp || audioMetrics.timestampNanos || audioMetrics.timestampNs || 0,
                timeunit: "NANOSECONDS"
            }            
        }).value();
    }

    function updatePPMLevelsForGraph ($graph, $pluginResults, $outputID)
    {
        var graphRootAsStr = $graph.root.toString();
        var ppmLevelsForGraph = (_ppmLevelsForGraph[graphRootAsStr] = _ppmLevelsForGraph[graphRootAsStr] || Maps.newConcurrentMap());

        $pluginResults = Objects.toJS($pluginResults, true);
        for (var key in $pluginResults)
        {
            var audioMetrics = $pluginResults[key];
            key = $outputID || key;
            ppmLevelsForGraph.put(key, makeAudioLevelsObject(key, audioMetrics));
        }
    }

    function getAudioLevelsForCaptureSessions ()
    {
        var allAudioLevels = {};
        for (var key in _activeCaptureGraphs)
        {
            var graph = _activeCaptureGraphs[key];
            if (!graph)
            {
                LOGGER.warning("Graph for " + key + " is not valid! (WTF?)");
                continue;
            }

            var audioLevelsForGraph = getPPMLevelsFromGraph(graph);
            var hasOneAudioLevel = _(audioLevelsForGraph).size() === 1;
            allAudioLevels[key] = hasOneAudioLevel ? _(audioLevelsForGraph).values().value()[0] : audioLevelsForGraph;
        }
        return allAudioLevels;
    }

    function setDevices ($someDevices)
    {
        _devices = Objects.toJS($someDevices);
    }

    function reset ()
    {
        _devices = [];
        _futureDevicePPMSessions = {};
        _activeCaptureGraphs = {};
        _ppmLevelsForGraph = {};
    }

    function setUpPPMForCaptureSessionGraph ($graph)
    {
        _(Objects.toJS($graph.nodes())).each(function ($node) {
            if (!(($node.type() === Processor.TYPE.WRITE || $node.type() === Processor.TYPE.INTEGRATION))) return;
            var outputID = $node.pluginID() + ":" + $node.id();

            if ($node.config().containsKey("id"))
            {
                outputID = $node.config().get("id");
            }

            _(Objects.toJS($node.parents())).each(function ($parentNode) {
                addPPMToNode($graph, $parentNode, outputID);
            });
        });

        return $graph;
    }

    function addPPMToNode ($graph, $node, $outputID)
    {
        var module = mediaProxy.getCaptureModule();
        var plugin = module.getPlugin($node.pluginID());

        if (!plugin.getOutputInterests().contains(Element.TYPE.AUDIO_FORMAT.name()) ||
            !$node.output(Element.TYPE.AUDIO_FORMAT).isPresent())
        {
            LOGGER.info("Not hooking up node #" + $node.id() + " ( a " + $node.pluginID() + ") because it's not producing an audio output.");
            return;
        }

        /*
         * If PPM node is already attached to this audio node, just
         * reuse the existing PPM node.
         */
        var ppmNode = _(Objects.toJS($node.children())).find(function ($node) {
            return $node.pluginID().equals(AUDIO_LEVEL_PLUGIN);
        });

        if (!ppmNode)
        {
            ppmNode = $graph.add(AUDIO_LEVEL_PLUGIN)
                            ['linkParents(java.lang.Iterable)']([$node])
                            .configure(PPM_ONLY_CONFIG);
        }

        ppmNode.invokeService(ON_AUDIO_LEVEL_UPDATE_PLUGIN_SERVICE, {
            callback: Objects.toConsumer(function ($data) {
                updatePPMLevelsForGraph($graph, $data, $outputID);
            })
        });
    }

    function addCaptureSession ($req, $graph, $stopPPMSessionsForCaptureDevices)
    {
        setUpPPMForCaptureSessionGraph($graph);
        _activeCaptureGraphs[$req.mediaPackageID] = $graph;

        if ($stopPPMSessionsForCaptureDevices)
        {
            var helper = new TemplateHelper($graph);
            Futures.allAsList(_(Objects.toJS(helper.getSources())).filter(function ($source) {
                return $source.type === MediaTypes.Device.AUDIO;
            }).map(function ($source) {
                LOGGER.info("Stopping PPM monitoring session for [" + $source.sourceId + "] because a new capture session is starting.");
                return removeDevice($source.sourceId);
            }).value()).get();
        }

        return $graph;
    }

    function addDevice ($device)
    {
        if (!($device.name in _futureDevicePPMSessions))
        {
            try
            {
                LOGGER.info("PPM monitoring session for [" + $device.name + "] is starting up.");
                var ppmRoot = new File(PPM_DIR, "ppm-" + System.currentTimeMillis() + "-" + StringUtilities.createSafeID($device.name));
                var mediaProcessingModule = mediaProxy.getCaptureModule();
                var ppmMonitorGraph = mediaProcessingModule.getGraphBuilder().get(ppmRoot);
                ppmMonitorGraph.make();

                var audioNode = ppmMonitorGraph.add($device.pluginID)
                                                ['linkParents(java.lang.Iterable)']([ppmMonitorGraph.node(0).get()])
                                                .configure({ device: $device.createConfiguration() });

                addPPMToNode(ppmMonitorGraph, audioNode);

                var futureSession = ppmMonitorGraph.startCapture();
                futureSession.addListener(function() {
                    var session = futureSession.get();
                    LOGGER.info("PPM monitoring session for [" + $device.name + "] started up.");

                    session.onSessionTerminated(Objects.toConsumer(function () {
                        LOGGER.info("PPM monitoring session for [" + $device.name + "] terminated.");

                        if (ppmRoot.exists())
                        {
                            try
                            {
                                FileUtil.deleteDir(ppmRoot);
                            }
                            catch (e)
                            {
                                LOGGER.info("Failed to delete PPM session directory: " + e);
                            }
                        }
                        
                        delete _futureDevicePPMSessions[$device.name];
                    }));
                }, EXECUTOR);

                /*
                 * Send an incident report if the PPM takes longer than expected to start up
                 */
                EXECUTOR.submit(new FutureStatusWatcher(futureSession, Duration.of(WATCHER_TIMEOUT_SECONDS, ChronoUnit.SECONDS), 5, Objects.toRunnable(function () {
                    var incident = new IncidentReport("PPM monitoring session for device [" + $device.name + "] is taking longer than expected to start up");
                    incident.setInternal(true);
                    facade.sendNotification(facade.HANDLE_EXCEPTION, incident);
                })));

                _futureDevicePPMSessions[$device.name] = futureSession;
            }
            catch (e)
            {
                facade.sendNotification(facade.HANDLE_EXCEPTION, new IncidentReport("PPM monitoring session for device [" + $device.name + "] failed to start up"));
            }
        }

        return _futureDevicePPMSessions[$device.name];
    }

    function removeDevice ($device)
    {
        var deviceId = $device.name || $device;
        if (!(deviceId in _futureDevicePPMSessions))
        {
            return Futures.immediateFuture(false);
        }

        LOGGER.info("PPM monitoring session for [" + deviceId + "] is shutting down.");
        var future = _futureDevicePPMSessions[deviceId].get().stop();
        
        /*
         * Send an incident report if the PPM takes longer than expected to start up
         */
        EXECUTOR.submit(new FutureStatusWatcher(future, Duration.of(WATCHER_TIMEOUT_SECONDS, ChronoUnit.SECONDS), 5, Objects.toRunnable(function () {
            var incident = new IncidentReport("PPM monitoring session for device [" + deviceId + "] is taking longer than expected to terminate");
            incident.setInternal(true);
            facade.sendNotification(facade.HANDLE_EXCEPTION, incident);
        })));

        return future;
    }

    function getAllMonitorableDevices ()
    {
        let filters = deviceFilters();        
        return _(Objects.toJS(_devices))
            .filter(filters.length > 0 ? device => !_(filters).find(filter => filter(device))
                                       : () => true)
            .value();
    }

    function addAllMonitorableDevices ()
    {
        _(getAllMonitorableDevices()).each($device => {
            var graph = addDevice($device);
            if (!graph)
            {
                LOGGER.warning("PPM graph for " + $device + " is not valid! (WTF?)");
                return;
            }
        });
    }

    function removeAllMonitorableDevices ()
    {
        _(getAllMonitorableDevices()).each(removeDevice);
    }

    function getAudioLevels ()
    {
        var levels = _({}).extend(getAudioLevelsForCaptureSessions()).value();

        _(getActiveDevicePPMSessions()).each($session => {
            _(levels).extend(getPPMLevelsFromGraph($session.graph)).value();
        });
        return levels;
    }

    function getActiveDevicePPMSessions ()
    {
        return _(_futureDevicePPMSessions)
            .filter($session => $session.isDone())
            .map($session => $session.get())
            .value();
    }

    return {
        getAudioLevels: getAudioLevels,
        setDevices: setDevices,
        addCaptureSession: addCaptureSession,
        addDevice: addDevice,
        removeDevice: removeDevice,
        getAllMonitorableDevices: getAllMonitorableDevices,
        addAllMonitorableDevices: addAllMonitorableDevices,
        removeAllMonitorableDevices: removeAllMonitorableDevices,
        reset: reset
    };
})();

function errorToIncident (err, message)
{
    var exception = err ? (err.javaException || new java.lang.Exception(err.message)) : null;
    message = message || err.message;
    var incident = new IncidentReport(message, exception);
    incident.setInternal(true);
    return incident;
}

function ppmLevelsToEventBody (levels)
{
    if (!levels || typeof levels !== "object")
    {
        return levels;
    }
    else if ("DB" in levels)
    {
        return levels.DB;
    }
    
    return _(levels)
        .mapValues(ppmLevelsToEventBody)
        .value();
}

function startPollingThread ()
{
    function hasConnections ()
    {
        return true;
    }

    function dispatchPPMEvent (levels)
    {
        facade.sendNotification(facade.EMIT_EVENT, new CatturaEvent(
            "AudioLevelsEvent",
            ppmLevelsToEventBody(levels)
        ));
    }

    var runner = new Runner.Atomic("PPM poller", true);
    runner.register(Objects.toRunnerListener(function () {
        try
        {
            /*
             * Return immediately if there are no more connections.
             */
            if (hasConnections())
            {
                /*
                 * Add all monitorable devices to the PPM meter.
                 */
                PPM.addAllMonitorableDevices();

                /*
                 * Ask audio devices for their sound levels
                 */
                var levels = PPM.getAudioLevels();
                if (_(levels).isEmpty())
                {
                    return;
                }

                dispatchPPMEvent(levels);
            }
            else
            {
                PPM.removeAllMonitorableDevices();
            }
        }
        catch (e)
        {
            facade.sendNotification(facade.HANDLE_EXCEPTION, errorToIncident(e));
        }
        finally
        {
            /*
             * Sleep for a bit
             */
            java.lang.Thread.sleep(POLL_TIMEOUT_MS);
        }
    }));

    var startPolling = sync(function () {
        if (!this.isRunningOrWaiting())
        {
            this.requestStartup();
        }
    }).bind(runner);
    startPolling();
}

on({
    AudioDevicesDetectedEvent: function ($event)
    {
        if ($event.isFinishedDetecting())
        {
            PPM.setDevices($event.devices);
        }
    },

    MediaProcessingModuleLaunchEvent: function ()
    {
        PPM.reset();
    },

    /*
     * Invoked when a new CaptureRequest is made (ie user start capture or scheduled event).
     * Will not fire for PPM or live device preview session for feedback.
     */
    CaptureSessionConstructionEvent: function ($event)
    {
        var req = $event.getRequest();
        var graph = $event.getGraph();
        PPM.addCaptureSession(req, graph, STOP_PPM_SESSIONS_FOR_CAPTURE_SESSIONS);
    },
    
    WebServerEvent: function ()
    {
        var API = require("../../web/cattura.web.api.APIPlugin/apiHelper")(plugin);
        API({
            audioLevels: {
                title: "Audio Levels API",
                description: "Methods for monitoring and managing audio level readings from devices and recordings",
                version: 2,
                methods: {
                    get: {
                        description: "Return all audio level data for devices and currently running recordings.",
                        verb: "get",
                        endpoint: "audioLevels",
                        callback: function() {
                            return PPM.getAudioLevels();
                        }
                    }
                }
            }
        });
    }
});

/*
 * If audio devices have been detected, then populate the PPM.
 */
if (devices.getDetectionProgress() >= 1)
{
    PPM.setDevices(devices.getAudioDeviceList());
}

plugin.onLoad(startPollingThread);
